#!/usr/bin/env python3

#This file imports the members defined in constants.h into a separated python module
# (from the "flat" module CLICopti.CLICopti generated by SWIG)

from CLICopti.CLICopti import \
    speed_of_light, \
    mu0, \
    electron_charge