#!/usr/bin/env python3

#Copy the content of the autogenerated CLICopti.py
# into the module namespace

#TODO: Separate the CellBase, AccelStructure, and CellParams namespaces
#from CLICopti.CLICopti import *

#Import sub-libraries so that they are automatically available
import CLICopti.CellBase
import CLICopti.CellParams
import CLICopti.Constants
import CLICopti.RFStructure